import java.awt.Point;
import java.awt.Color;
import java.util.LinkedList;
import java.awt.geom.*;
import H.*;


class CoordinateSystem extends Drawable
{    
    
    protected double m_xmin = Double.MAX_VALUE;
    protected double m_xmax = -Double.MAX_VALUE;
    protected double m_ymin = Double.MAX_VALUE;
    protected double m_ymax = -Double.MAX_VALUE;
    protected double m_xdelta;
    protected double m_ydelta;
    
    protected double m_theta;
    
    private LinkedList csObjects = new LinkedList();
    
    private Point tempPoint = new Point(0,0);
    
    // inner class for Lines
    
    class CS_Line
    {
        public double x1, y1, x2, y2;
        public Color c;
        
        public CS_Line(double x1_, double y1_, 
                       double x2_, double y2_, Color c_)
        {
            x1 = x1_; y1 = y1_; x2 = x2_; y2 = y2_; c = c_;
        }
    }
    
    // overridden from class Drawable
    // so that we can clear the screenObjects list first
    protected void register()
    {
        H.pl("Registering the cs object");
        //Drawable.screenObjects = new LinkedList();
        //Drawable.screenObjects.add(this);
    }
    
    public CoordinateSystem()
    {
        setCS( -1.0, 10.0, -1.0, 10.0 );
    }
    
    public CoordinateSystem(double xmin, double xmax, double ymin, double ymax)
    {
        setCS( xmin,  xmax,  ymin,  ymax);
    }
    
    public void setCS(double xmin, double xmax, double ymin, double ymax)
    {
        m_xmin = xmin;
        m_xmax = xmax;
        m_ymin = ymin;
        m_ymax = ymax;
        m_xdelta = m_xmax - m_xmin;
        m_ydelta = m_ymax - m_ymin;
        m_theta = 0.0;
        //at = new AffineTransform();
        isSaneCS();
    }
    
    public void squareCoordinates()
    {
        ;// should enlarge one dimension to square things
    }
    
    public boolean isSaneCS()
    {   
        if (m_xmin < m_xmax && m_ymin < m_ymax)
            return true;
        else
        {    
            H.pl("Warning: xmin >= xmax or ymin >= ymax in CoordinateSystem");
            H.pl("Values are (xmin,xmax,ymin,ymax) = (" + m_xmin + "," + m_xmax 
                 + "," + m_ymin + "," + m_ymax + ")");
            H.pl("Resetting to x in (-20,20)  y in (-20,20).\n");
            setCS(-20,20,-20,20);
        }
        return false;
    }
    
    /*
     * p gets the output point 
     * in pixel units equivalent to x,y
     */
    public void convertPoint(double x, double y, Point p)
    {        
        double xpercent = (x-m_xmin)/m_xdelta;
        double ypercent = (y-m_ymin)/m_ydelta;
        
        // convert to pixel values
        
        p.x = (int)((xpercent*width)+0.5);
        p.y = (height-1)-(int)((ypercent*height)+0.5); // top-bottom reversal
    }
        
    
    /*
     * draws the point in the logical (not pixel) coordinate system
     */
    public void drawPoint(double x, double y)
    {
//        H.pl("x,y are " + tempPoint.x + "," + tempPoint.y );
//        H.pl("Values are (xmin,xmax,ymin,ymax) = (" + m_xmin + "," + m_xmax 
//                 + "," + m_ymin + "," + m_ymax + ")");
        // if we're out of the range, don't do anything
        if (x < m_xmin || x > m_xmax || y < m_ymin || y > m_ymax)
            return;
            
        convertPoint(x,y,tempPoint);
                    
        int w = 0; // how big to draw a single point
        //H.pl("x,y are " + tempPoint.x + "," + tempPoint.y );
        g.fillRect(tempPoint.x-w,tempPoint.y-w,2*w+1,2*w+1);
    }   
    
    public void fillRect(double x, double y, double w, double h)
    {
        // if we're offscreen, no need to draw
        if (x+w < m_xmin || x > m_xmax || y+h < m_ymin || y > m_ymax)
            return;
            
        // we will take x,y to be the upper right corner
        convertPoint(x,y,tempPoint); 
        
        // convert width and height now...
        
        w *= width/m_xdelta;
        h *= height/m_ydelta;
                    
        g.fillRect(tempPoint.x,tempPoint.y,(int)(w+0.5),(int)(h+0.5));
    } 
    
    public void addLine(double x1, double y1, double x2, double y2, Color c)
    {
        CS_Line csl = new CS_Line(x1,y1,x2,y2,c);
        csObjects.add(csl);
    }
    
    public void drawLine(double x1, double y1, double x2, double y2, Color c)
    {
        // we draw the line even if both points are off screen, 
        // because part of the line might be onscreen!
        int px1,py1;
        convertPoint(x1,y1,tempPoint);
        px1 = tempPoint.x;
        py1 = tempPoint.y;
        convertPoint(x2,y2,tempPoint);
        
        g.setColor(c);
        g.drawLine(px1,py1,tempPoint.x,tempPoint.y);
    }
    
    public void draw()
    {
        // print the window bounds
        H.pl("Coordinates' bounds: " + 
             "(" + m_xmin + " < x < " + m_xmax + ")   " +
             "(" + m_ymin + " < y < " + m_ymax + ")");
        
        // draw the axes
        drawLine(m_xmin,0,m_xmax,0, Color.green);
        drawLine(0,m_ymin,0,m_ymax, Color.green);
        
        // draw the lines the user added
        for (int i=0 ; i<csObjects.size() ; ++i)
        {
            CS_Line csl = (CS_Line)(csObjects.get(i));
            drawLine(csl.x1,csl.y1,csl.x2,csl.y2,csl.c);   
        }
    }

}
