import java.awt.Color;
import H.*;

class LifeGrid extends Grid
{
    private int[][] cells1;
    private int[][] cells2;
    private Color aliveColor = Color.black;  // default colors
    private Color emptyColor = Color.white;
    private CS5App m_linkToCode;
    private int counter = 0;
    
    public LifeGrid(int nrows, int ncols, CS5App linkToCode)
    {
        super(nrows,ncols);
        m_linkToCode = linkToCode;
        cells1 = new int[nrows][ncols];
        cells2 = new int[nrows][ncols];
        initializeAll();
    }
    
    public void nextStep()
    {
        // call the code in CS5AppGr
        //  pass as arguments the cells
        //  if counter is even, cells1 has the current state of the grid
        //  if counter is odd, cells2 has the current state of the grid
        if (counter%2 == 0)
        {
            //m_linkToCode.updateCells(cells1,cells2);
            setColors(cells2);
        }
        else
        {
            //m_linkToCode.updateCells(cells2,cells1);
            setColors(cells1);
        }
        
        ++counter;
        
    }
    
    public void setAliveColor(Color c)
    {
        this.aliveColor = c;
    }
    
    public void setEmptyColor(Color c)
    {
        this.emptyColor = c;
    }    
    
    protected void initializeAll()
    {
        for (int row=0 ; row<m_nrows ; ++row)
        {
            for (int col=0 ; col<m_ncols ; ++col)
            {  
                cells2[row][col] = 0;
                if (Math.random() < 0.5)
                    cells1[row][col] = 0;
                else 
                    cells1[row][col] = 1;
                    
                // set border to all empty
                if (col == 0 || col == m_ncols-1 || row == 0 || row == m_nrows-1) 
                    cells1[row][col] = 0;
            }
        }
        setColors(cells1);
    }
    
    protected void setColors(int[][] A)
    {
        // what if A is the wrong size?
        // we'll print a warning and set as much as possible
        if (A == null || A[0] == null)
        {
            H.pl("LifeGrid::setColors called with a null array. Returning...");
            return;
        }
        
        int lastrow = m_nrows;
        int lastcol = m_ncols;
        
        if (A.length != m_nrows || A[0].length != m_ncols)
        {
            H.pl("LifeGrid::setColors called with a missized array. (warning)");
            lastrow = A.length < m_nrows ? A.length : m_nrows;
            lastcol = A[0].length < m_ncols ? A[0].length : m_ncols;
        }
        
        for (int row=0 ; row<lastrow ; ++row)
        {
            for (int col=0 ; col<lastcol ; ++col)
            {
                m_colors[row][col] = A[row][col] == 1 ? aliveColor : emptyColor;
            }
        }
    }
}