import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.Vector;
import java.util.Collections;
import javax.swing.*;
import H.*;

class Pose
implements Comparable
{
	public double x;
	public double y;
	public double theta;  // in radians
	public double p;
	
	public double ir_ne = 0.0;
	public double ir_n  = 0.0;
	public double ir_nw = 0.0;
	
	public Line2D.Double ir_ne_l = new Line2D.Double();
	public Line2D.Double ir_n_l  = new Line2D.Double();
	public Line2D.Double ir_nw_l = new Line2D.Double();
	
	public Pose(double x, double y, double theta, double p)
	{
		this.x = x; this.y = y; this.theta = theta; this.p = p;
	}
	
	public int compareTo(Object o)
	{
		if (o instanceof Pose)
		{
			if (this.p < ((Pose)o).p)
				return 1;
			else if (this.p == ((Pose)o).p)
				return 0;
			else
				return -1;
		}
		else
			return 0;
	}
	
}

class PFilterParams
{
	public double d_spread; // fraction of distance (0 to 1)
	public double theta_spread; // fraction of angle (0 to 1)
	public double theta_jitter;
	
	public PFilterParams()
	{
		d_spread = PFilter.TRANS_ERR_WIDTH;
		theta_spread = PFilter.ROTERR_WIDTH;
		theta_jitter = Math.toRadians(PFilter.JITTER);
	}
}

class PFilter implements Drawable
{
    public Vector particles;   // bunch of estimates...
    public AffineTransform AT,savedAT;
    
    public int size; // number of particles
    public int radius = 5; // size of particles
    
    public int type;
    public static final int UNIFORM = 0;
    public static final int TRIANGLE = 1;
    public static final int GAUSSIAN = 2;
    
    public static  double JITTER;
    public static  double TRANS_ERR_WIDTH;
    public static  double ROTERR_WIDTH;
    
    public PFilterParams pfp;
    
    
    public PFilter(int size, int type,  PFilterParams pfp)
    {
    	this.type = type;
    	this.size = size;
    	this.pfp = pfp;

        initialize(this.size);
    }
    
    public PFilter(int size)
    {
		this.type = PFilter.UNIFORM;
		this.size = size;
		this.pfp = new PFilterParams();   
		initialize(this.size);
    }
    
    public void updateWithSensorValues(double ir_nw_sensed, double ir_n_sensed, double ir_ne_sensed, double wi_sensed)
    {
    	updateWithSensorValues(ir_nw_sensed,ir_n_sensed,ir_ne_sensed,wi_sensed, true);
    }
    
    // when boundsCheck is true points outsie the env. are culled
    public void updateWithSensorValues(double ir_nw_sensed, double ir_n_sensed, double ir_ne_sensed, 
    	double wi_sensed, boolean boundsCheck)
    {
    	
    	if (this.type == PFilter.UNIFORM)
    	{
	        int listlen = particles.size();
	        Pose p = null;
	        for (int i=0 ; i<listlen ; ++i)
	        {
	        	p = (Pose)(particles.get(i));
	        	// update p...
	        	
	        	// for now, we will allow values within 50% of either side
	        	// of ir_n
	        	
	        	// so, we have to get the ir_n distance!
	        	// remember that ir_n is offset from the pose, which
	        	// estimates the robot's center...
	        	
	        	//if (i<1) { 
	        	
	        	if (boundsCheck && !Geo.inEnvironment(p.x,p.y))
	        	{
	        		p.p = 0.0;
	        		continue;
	        	}
	        	
	        	double irt = p.theta /* already in radians */, irx = 0.0, iry = 0.0, alpha = Math.atan2(10,40);
			    
			    irx = p.x+41.23*Math.cos(irt+alpha); iry = p.y+41.23*Math.sin(irt+alpha); 
			    double ir_ne = Geo.distToNearestWall(irx,iry,irt+0.52);
				
			    irx = p.x+40.0*Math.cos(irt); iry = p.y+40.0*Math.sin(irt); 
			    double ir_n = Geo.distToNearestWall(irx,iry,irt); // add true flag, if you like
			    //H.pl("irx, iry, irt (deg) are " + irx + " " + iry + " " + irt + " (" + Math.toDegrees(irt) + ")");
				
			    irx = p.x+41.23*Math.cos(irt-alpha); iry = p.y+41.23*Math.sin(irt-alpha); 
			    double ir_nw = Geo.distToNearestWall(irx,iry,irt-0/52);
			    
			    double wi = CS5App.envMap.getWiSig(p.x, p.y);
			    

				// update the probability of p, set to null if it doesn't work
				
				if (Math.abs(wi - wi_sensed) > 10)
				{
					p.p /= 10;
				}
				
				
				// straight ahead
                if (ir_n < ER1.IR_N_TH && ir_n_sensed > ER1.IR_N_TH + 10)
				{
				    p.p /= 10;
				}
				
                if (ir_n > ER1.IR_N_TH + 10 && ir_n_sensed < ER1.IR_N_TH)
				{
				    p.p /= 10;
				}
				
				// to the left
				if (ir_nw < ER1.IR_NW_TH && ir_nw_sensed > ER1.IR_NW_TH + 10)
				{
				    p.p /= 10;
				}
				
                if (ir_nw > ER1.IR_NW_TH + 10 && ir_nw_sensed < ER1.IR_NW_TH)
				{
				    p.p /= 10;
				}
				
				// to the right
				if (ir_ne < ER1.IR_NE_TH && ir_ne_sensed > ER1.IR_NE_TH + 10)
				{
				    p.p /= 10;
				}
				
                if (ir_ne > ER1.IR_NE_TH + 10 && ir_ne_sensed < ER1.IR_NE_TH)
				{
				    p.p /= 10;
				}
				
				/*
				if (ir_n > 1.5*ir_n_sensed || ir_n < .5*ir_n_sensed)
				{
					p.p /= 100.0;
				}
				*/
				
				p.ir_n = ir_n;
				//if (i < 1) H.pl("i = " + i + "  p.ir_n is " + p.ir_n);
				
				//}
	        }
    	}	
    	else if (this.type == PFilter.TRIANGLE)
    	{
;
    	}
    	
    	else if (this.type == PFilter.GAUSSIAN)
    	{
    		;
    	}
    	
    	else
    	{
    		H.pl("Unrecognized PFilter motion model: " + this.type);
    	}
    	
    	// next, we need to resample 
    	
    	int listlen = particles.size();
    	Collections.sort(particles);
        Pose p = null;
        Vector newv = new Vector();
        double sum = 0.0;
        for (int i=0 ; i<listlen ; ++i)
        {
        	p = (Pose)(particles.get(i));
        	H.p(p.p + " ");
        	sum += p.p; // add up
        }
        H.pl();
       
       
        // see notes for resampling technique
        double width = sum/listlen;
        double runningSum = 0.0;
        int currentParticle = -1;
        
        //H.p(".r. ");  // message that we're resampling
        
        for (int i=0 ; i<listlen ; ++i) // place in the new vector
        {
        	
        	while (runningSum < width/2 + i*width && currentParticle<listlen)
        	{
        		++currentParticle;
        		p = (Pose)(particles.get(currentParticle));
        		runningSum += p.p;
        	}
        	//p = (Pose)particles.get(i);
        	//if (p.p >= .1)
        	//{
	        	Pose pnew = new Pose(p.x,p.y,p.theta,1.0);
	        	pnew.ir_ne = p.ir_ne;
	        	pnew.ir_n = p.ir_n;
	        	pnew.ir_nw = p.ir_nw;
	        	newv.add(pnew);
        	//}
        	
        	//H.p(" " + i);
        	
        }
        /*
        // Add in random for the rest
        int diff = listlen - newv.size();
        H.pl("diff: " + diff);
        for (int i = 0; i < .75 * diff; ++i) 
        {
        	p = (Pose)particles.get(i);
        	double x = p.x + (5 - Math.random() * 10);
        	double y = p.y + (5 - Math.random() * 10);
        	double th = Math.random() * 360;
        	while (!Geo.inEnvironment(x, y))
        	{
        		x = Math.random() * CS5App.envMap.env_max_x;
        		y = Math.random() * CS5App.envMap.env_max_y;
        		H.pl("x: " + x + "y: " + y);
        	}
        	Pose pnew = new Pose(x, y, th, 1.0);
        	newv.add(pnew);
        }
        
        for (int i = 0; i < .25 * diff; ++i) 
        {
        	double x = CS5App.envMap.env_max_x - Math.random() * CS5App.envMap.size_x;
        	double y = CS5App.envMap.env_max_y - Math.random() * CS5App.envMap.size_y;
        	double th = Math.random() * 360;
        	while (!Geo.inEnvironment(x, y))
        	{
        		x = Math.random() * CS5App.envMap.env_max_x;
        		y = Math.random() * CS5App.envMap.env_max_y;
        		H.pl("x: " + x + "y: " + y);
        	}
        	Pose pnew = new Pose(x, y, th, 1.0);
        	newv.add(pnew);
        }
		*/
        particles = newv; // go to new vector for our next sample
        
    }
    
    // turns dth1 (in the direction of the final displacement vector
    // moves dist
    // turns dth2  (i.e. by that angular amount)
    public void updateWithMotionModel(double dth1, double dist, double dth2)
    {
    	if (this.type == PFilter.UNIFORM)
    	{
	        int listlen = particles.size();
	        Pose p = null;
	        //H.pl("pfp.d_spread is " + pfp.d_spread);
	        for (int i=0 ; i<listlen ; ++i)
	        {
	        	p = (Pose)(particles.get(i));
	        	// update p...
	        	double guess_th1  = dth1*(1.0+pfp.theta_spread*(Math.random()*2-1));
	        	double guess_dist = dist*(1.0+pfp.d_spread*(Math.random()*2-1));
	        	double guess_th2  = dth2*(1.0+pfp.theta_spread*(Math.random()*2-1));
	        	double theta_jitter = (Math.random()*2-1)*pfp.theta_jitter;
	        	
	        	double guess_th_move = p.theta + guess_th1;

	        	p.x += guess_dist*Math.cos(guess_th_move);
	        	p.y += guess_dist*Math.sin(guess_th_move);
	        	p.theta = guess_th_move + guess_th2 + // the next term for the jitter
	        	            theta_jitter;
	        	//H.p(dist + " " + guess_dist + "|");
	        }
    	}
    	
    	else if (this.type == PFilter.TRIANGLE)
    	{
    		;
    	}
    	
    	else if (this.type == PFilter.GAUSSIAN)
    	{
    		;
    	}
    	
    	else
    	{
    		H.pl("Unrecognized PFilter motion model: " + this.type);
    	}
    }
    
    public void initialize(int size)
    {
        initialize(size,ER1.ALL_SAME,ER1.ALL_SAME_DIR);
    }
    
    public void initialize(int size, boolean allSame, boolean allSameDir)
    {
        H.pl("Point A");
    	this.size = size;
        AT = null;
        particles = new Vector();
        
    	double x,y,t,p;
    	
    	particles.add(new Pose(0,0,0.0,1.0)); // at robot's starting point
        particles.add(new Pose(0,0,0.0,1.0)); // at robot's starting point
        particles.add(new Pose(0,0,0.0,1.0)); // at robot's starting point
        particles.add(new Pose(0,0,0.0,1.0)); // at robot's starting point
        particles.add(new Pose(0,0,0.0,1.0)); // at robot's starting point
    	int i=5;
    	
    	// special case...
            if (allSame && !allSameDir) // special case
            {
                EnvMap.env_max_x = 200;  EnvMap.env_min_x = -200;
                EnvMap.env_max_y = 100;  EnvMap.env_min_y = -100;
            }
    	
    	while (i<size)
    	{
            // H.pl("Point B i = " + i);
    	    if (!allSameDir)
    	    {
        		x = (EnvMap.env_max_x - EnvMap.env_min_x)*Math.random() + EnvMap.env_min_x;
        		y = (EnvMap.env_max_y - EnvMap.env_min_y)*Math.random() + EnvMap.env_min_y;
    
        		
        		if (Geo.inEnvironment(x,y))
        		{
        		
        		    t = Math.random()*2.0*Math.PI;
        		    particles.add(new Pose(x,y,t,1.0));
        		    ++i;
        		    //H.p(" " + i);
        		}
    		}
    		else if (!allSame)  // just all same direction
    		{
                x = (EnvMap.env_max_x - EnvMap.env_min_x)*Math.random() + EnvMap.env_min_x;
                y = (EnvMap.env_max_y - EnvMap.env_min_y)*Math.random() + EnvMap.env_min_y;
    
                
                if (Geo.inEnvironment(x,y))
                {
                    particles.add(new Pose(x,y,0.0,1.0));
                    ++i;
                    //H.p(" " + i);
                }
    		}
    		else // all the same point!
    		{
                particles.add(new Pose(0,0,0.0,1.0));
                ++i;
    		}
    		
    	}

    	/*
    	for ( i=1 ; i<size ; ++i)
    	{
    		x = 2*xhw*Math.random() - xhw;
    		y = 2*yhw*Math.random() - yhw;
    		t = Math.random()*2.0*Math.PI;
    		particles.add(new Pose(x,y,t,1.0));
    	}
    	*/
    }
    
    public void draw(Graphics2D g)
    {
    	/*
    	savedAT = g.getTransform();
        currentAT.translate(deltaX+currentWidth/2,deltaY+currentHeight/2); // center
        AT = AffineTransform.getScaleInstance(zoom,zoom);
        g.transform(currentAT);
        g.transform(AT);
        */

        int listlen = particles.size();
        Pose p = null;
        g.setColor(Color.lightGray);
        for (int i=(int).25*listlen ; i<listlen ; ++i)
        {
        	p = (Pose)(particles.get(i));
        	if (p.p > 0)
        	{
	        	g.fill(new Ellipse2D.Double(p.x-radius,p.y-radius,2*radius,2*radius));
	        	g.draw(new Line2D.Double(p.x,p.y,p.x+2*radius*Math.cos(p.theta),p.y+2*radius*Math.sin(p.theta)));
	        	/*
	        	g.setColor(Color.orange);
	        	double c = Math.cos(p.theta);
	        	double s = Math.sin(p.theta);
                p.ir_n_l.setLine(p.x + 40*c,p.y + 40*s,p.x+(40+p.ir_n)*c,p.y+(40+p.ir_n)*s);
                g.draw(p.ir_n_l);
	        	*/
	        	
	        	/*
	        	if (i < 1)
	        	{ 
		        	H.pl("(Draw) i = " + i + "  p.ir_n is " + p.ir_n);
		        	if (i == 0) g.setColor(Color.orange);
		        	if (i == 1) g.setColor(Color.red);
		        	if (i == 2) g.setColor(Color.black);
		        	//ir_nw_l.setLine(40,-10,40+CS5App.er1.ir_nw,-10); // could include an angle, too...
		        	double c = Math.cos(p.theta);
		        	double s = Math.sin(p.theta);
	                p.ir_n_l.setLine(p.x + 40*c,p.y + 40*s,p.x+(40+p.ir_n)*c,p.y+(40+p.ir_n)*s);
	                g.draw(p.ir_n_l);
	                //ir_ne.setLine(40,10,40+CS5App.er1.ir_ne,+10);
	            }
	            */
	        }
        }
        g.setColor(Color.pink);
        for (int i=0 ; i<.25 * listlen ; ++i)
        {
        	p = (Pose)(particles.get(i));
        	if (p.p > 0)
        	{
	        	g.fill(new Ellipse2D.Double(p.x-radius,p.y-radius,2*radius,2*radius));
	        	g.draw(new Line2D.Double(p.x,p.y,p.x+2*radius*Math.cos(p.theta),p.y+2*radius*Math.sin(p.theta)));
	        }
	        //H.p("d");
	    }
        /*
        currentAT.setToIdentity();
        g.setTransform(savedAT);
        */
    }
    
    public void handleResizeEvent()
    {
    	;
    }
    
    
}

















    