import java.awt.*;
import java.awt.event.*;
import java.util.LinkedList;
import javax.swing.*;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.StreamTokenizer;
import H.*;

class Picobot extends Grid
{
    // a global made false by changing the env. or moving the robot
    // a user gets "credit" for finishing only if this is true
    private boolean allowedToFinish = true;
    
    // keeping track of progress
    private int allCellsRemaining;
    private int borderCellsRemaining;
    public static int totalCoverageMoves = -1;
    public static int perimCoverageMoves = -1;
        
    private int m_mapnum;
    private GridCell[][] grid;
    private static int defsize = 25;  // default size
    public static Color waterColor = Color.blue;
    public static Color aliveColor = Color.white;   
    public static Color robotColor = Color.green;   
    public static Color seenColor = Color.lightGray;   
    private int robotx;
    private int roboty;
    
    
    public static int statenow;
    static int nMoves = 0;          // number of moves the robot has made
    static int movesPerCheck = 50;   // number of moves between checks for completion


    static final int MAXSUR = 17;
    static final int MAXSTATE = 100;     
     // the rules 0 to MAXSTATE-1 are legal to specify
    static State[] state = new State[MAXSTATE+1];  // to be safe... 

    
    public Picobot()
    {
        /*
         * notice the dependence on 25!
         */
        super(defsize,defsize); // or whatever...
        parseRules("picobotRules");
        robotx = 0;
        roboty = 0;
        
        /*
         * here is where you can change the map --
         * just be sure to save (control-s) and recompile (F7)
         * before rerunning (F5)
         *
         * Available maps are 0-6.
         */
        changeMap(0); // change to the initial map
        
        // set up event handlers
        m_canvas.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseClicked(java.awt.event.MouseEvent e) {
                // by clicking in the world, you can't get "credit" for it
                allowedToFinish = false;
                int modifiers = 0;
                if (e.isControlDown()) modifiers = 1;
                else if (e.isShiftDown()) modifiers = 2;
                handleClick(e.getX(),e.getY(),modifiers);
            }
        });
    }
    
    void handleClick(int x, int y, int control)
    {
        
        H.pl("x,y are " + x + "," + y);
    
        int col = (x-m_xorig)/m_maxdim;
        int row = (y-m_yorig)/m_maxdim;
               
        H.pl("col is " + col + " and row is " + row);
        
        if (row >= 1 && row < defsize-1 && col >= 1 && col < defsize-1)
        {
          if (control==1) // option
          {
            grid[roboty][robotx].setColor(aliveColor);
            roboty = row; robotx = col;
            grid[roboty][robotx].setColor(robotColor);
          }
          else if (control == 2)  // shift
          {
            if (grid[row][col].getPebble())
            {
              grid[row][col].setPebble(false);
            }
            else
            {
              grid[row][col].setPebble(true);
            }
          }
          else // no button
          {
              if ( grid[row][col].getColor() == waterColor )
              {
                grid[row][col].setColor(aliveColor);
              }
              else
              {
                grid[row][col].setColor(waterColor);
              }
            }
        }
        // if there is a pebble, it gets removed and put back in the 
        // supply (infinite supply for now)
        
        /*
        else
        {
          System.out.println("Out of the grid's bounds");
        }
        */
        
        m_canvas.repaint();
    }


    
    public void nextStep()
    {
        // are we already finished?
        if (allCellsRemaining == 0 && allowedToFinish) {
            H.pl("This environment is complete.");
            OuterFrame.updater.off();
            return;
        }
        
        // decide which way to move the robot
        // and anything about the pebbles, then
        int nextMove = getNextMove();
        
        if (!moveRobot(nextMove))
        {
            OuterFrame.updater.off(); // it stopped! stop the simulation
        }
        else
        {
            ++totalCoverageMoves;
        }
        
        if (allCellsRemaining == 0 && allowedToFinish) 
        {
            H.pl("\n\n********************************************");
            H.pl(    "* Successful Total Coverage! " + 
                computeBuffer(totalCoverageMoves,5) + totalCoverageMoves +
                " moves   *");
            H.pl(    "********************************************\n\n");
        } 
    }
    
    public void draw()
    {
        if (grid != null && grid[0][0] != null) {
        for (int row=0 ; row<m_nrows ; ++row)
            for (int col=0 ; col<m_ncols ; ++col)
                m_colors[row][col] = grid[row][col].getColor();
        super.draw();
        }
    }
    
    public void changeMap(int mapnum)
    {
        int[][] map = null;
        
        switch (mapnum)
        {
            case 0: map = Map.map0; break;
            case 1: map = Map.map1; break;
            case 2: map = Map.map2; break;
            case 3: map = Map.map3; break;
            case 4: map = Map.map4; break;
            case 5: map = Map.map5; break;
            case 6: map = Map.map6; break;
            default: map = Map.map0; mapnum = 0; break;
        }
    
        m_mapnum = mapnum;
    
        m_nrows = map.length;
        m_ncols = map[0].length;
        
        grid = new GridCell[m_nrows][m_ncols];
    
        for (int row=0 ; row<m_nrows ; ++row)
        {
          for (int col=0 ; col<m_ncols ; ++col)
          {
            grid[row][col] = new GridCell();
            
            if (map[row][col] == 1) 
              grid[row][col].setColor(waterColor);
            else
              grid[row][col].setColor(aliveColor);
          }
        }
        startNewWorld();
        robotx = 0; roboty = 0; // reset robot
        setRandomRobot();
    }
    
    public boolean moveRobot(int move)
      {
        boolean retValue = true;  
        switch (move)
        {
          case Move.N:
            retValue = placeRobot(roboty-1,robotx,false,'N'); break;
          case Move.S:
            retValue = placeRobot(roboty+1,robotx,false,'S'); break;
          case Move.E:
            retValue = placeRobot(roboty,robotx+1,false,'E'); break;
          case Move.W:
            retValue = placeRobot(roboty,robotx-1,false,'W'); break;   
          case Move.X:
            break;
            // drop, pickup
          default:
            H.out.println("Unrecognized move #" + move);
        }
        return retValue;
      }


    public void startNewWorld()
    {
        statenow = 0;
        totalCoverageMoves = -1;
        perimCoverageMoves = -1;
        // global that allows them to finish
        allowedToFinish = true;
        // it's made false by changing the environment or
        // relocating the robot -- basically by using the mouse on
        // the grid
        
        // count number of spaces with blue (border) 8-neighbors
        // and count total number of white (unrepaired) spaces
        // simply decrement these counts as appropriate, and then
        // when 0, give credit as long as "allowedToFinish" is true
        
        borderCellsRemaining = 0;
        allCellsRemaining = 0;
        
        for (int row=0 ; row<m_nrows ; ++row)
        {
          for (int col=0 ; col<m_ncols ; ++col)
          {
            if (grid[row][col].getColor() == waterColor)
            {
              ;
            }
            else
            {
              ++allCellsRemaining;
              
              // check all eight neighbors...
              boolean blueneighbor = false;
              for (int k=-1 ; k<=1 ; ++k)
                for (int l=-1 ; l<=1 ; ++l)
                  if (grid[row+k][col+l].getColor() == waterColor)
                    blueneighbor = true;
              
              if (blueneighbor)
                ++borderCellsRemaining;
            }
          }
        }
        
    }

    
    public void setRandomRobot()
    {
        int heldrow=-1, heldcol=-1;
        int count = 0;
        // famous uniform randomizer on the fly
        for (int row=0 ; row<m_nrows ; ++row)
        {
          for (int col=0 ; col<m_ncols ; ++col)
          {
            if (grid[row][col].getColor() == aliveColor)
            {
              ++count;
              if (Math.random() < 1.0/count)
              {
                // switch to new one
                heldrow = row;
                heldcol = col;
              }
            }
          }
        }
        placeRobot(heldrow,heldcol,true,'X');
    }
    
    public boolean placeRobot(int ry, int rx, boolean reset, char dir)
    {
        if (rx >= 1 && ry >= 1 && rx < m_ncols-1 && ry < m_nrows-1)
        {
           // check if it is a nerve cell...
           if (grid[ry][rx].getColor() == waterColor)
           {
               System.out.println(" You have attempted to move " + dir + 
                                  " onto an obstacle!\n"); 
                 // at " +  "(" + rx + ", " + ry + ")!\n" +
                 //         " The robot will NOT be moved.\n");
                          
                 //System.out.println("The grid is\n" + grid);              
                 grid[roboty][robotx].setColor(robotColor);
                 return false;
           }
             
           // make old robot square visited color
           if (roboty > 0 && robotx > 0)  // not the first time!
           {
             if (reset)
                 grid[roboty][robotx].setColor(aliveColor);
             else
                 grid[roboty][robotx].setColor(seenColor);
           }
            
           // update robot position 
           robotx = rx; roboty = ry;
           
           // make new robot square robot color
           // check the previous color of the ry, rx square
           // and its neighbors to see if the task is complete
           if (grid[ry][rx].getColor() == aliveColor)
           {
              --allCellsRemaining;
             
              // check all eight neighbors...
              boolean blueneighbor = false;
              for (int k=-1 ; k<=1 ; ++k)
                for (int l=-1 ; l<=1 ; ++l)
                  if (grid[ry+k][rx+l].getColor() == waterColor)
                  {  blueneighbor = true; l=k=2; }
              
              if (blueneighbor)
                --borderCellsRemaining;                            
           }
           
           grid[roboty][robotx].setColor(robotColor); 
           
        }
        else
        {
             System.out.println(" You have attempted to move " + dir + 
                                      " onto an obstacle!\n"); 
              return false;
          /*H.out.println("The requested robot coordinates of ("+
                         rx + "," + ry + ") are out of bounds.\n" + 
                        "The robot will not be moved.");*/
        }
        return true;

    }
    
      static public void printRules(State[] s)
  {
    for (int i=0 ; i<s.length ; ++i)
    {
      if ( s[i] == null ) continue;
      for (int j=0 ; j<s[i].trans.length ; ++j)
      {
        if ( s[i].trans[j] == null ) continue;
        H.out.println("(" + i + "," + j + ") -> (" +
          Move.print(s[i].trans[j].move) + "," + 
          s[i].trans[j].ns + ")");
      }
      if ( s[i].def == null ) continue;
      H.out.println("(" + i + ",*) -> (" +
          Move.print(s[i].def.move) + "," + 
          s[i].def.ns + ")");
    }
  }
  
  static public void checkRules(State[] s)
  {
    if ( s[0] == null )
    {
      H.out.println("Error: you have not defined a rule that " +
                    "handles state #0, the starting state!\n");
            System.exit(0);
    }
    for (int i=0 ; i<s.length ; ++i)
    {
      if ( s[i] == null ) continue;
      for (int j=0 ; j<s[i].trans.length ; ++j)
      {
        if ( s[i].trans[j] == null ) continue;
        int ns = s[i].trans[j].ns;
        if ( s[ns] == null )
        {
          H.out.println("Error: you have defined a rule that " +
                    "transitions to state #" + ns +"\n" +
                    "   -- But there is no rule for state #" + ns + "!\n");
          H.out.print("Here is the rule: ");
          H.out.println("(" + i + "," + j + ") -> (" +
          Move.print(s[i].trans[j].move) + "," + 
          s[i].trans[j].ns + ")");
                System.exit(0);
        }
      }
      if ( s[i].def == null ) continue;
      int ns2 = s[i].def.ns;
      
      if ( s[ns2] == null )
        {
          H.out.println("Error: you have defined a rule that " +
                    "transitions to state #" + ns2 +"\n" +
                    "  -- But there is no rule for state #" + ns2 + "!\n");
          H.out.print("Here is the rule: ");
          H.out.println("(" + i + ",*) -> (" +
          Move.print(s[i].def.move) + "," + 
          s[i].def.ns + ")");
                System.exit(0);
        }
    }
  }
  
  public String cb(int i)
  {
    return computeBuffer(i,2)+i;
  }
  
  public String computeBuffer(int i, int width)
  {
    String s = "";
    int neededWidth = 0;
    
    if (i==0)
    {
      neededWidth = width-1;
    }
    else
    {  
        int I = i;
        if (i < 0)
        {
          I = -i;
        }
        int numwidth = 1+(int)(Math.log(I+.5)/Math.log(10.0));
        neededWidth = width-numwidth;
        if (i<0)
          --neededWidth;
      }
      
      for (int j=0 ; j<neededWidth ; ++j)
      {
        s += " ";
      }
      
      return s;
  }
  
  // returns true if s.length() >= 1 and all of s's characters
  // are from the set news or NEWS
  public static boolean isCompass(String s)
  {
    if ( s.length() < 1 ) return false;
    for ( int i=0 ; i<s.length() ; ++i)
    {
      switch (s.charAt(i))
      {
        case 'n':
        case 'N':
        case 'e':
        case 'E':
        case 'w':
        case 'W':
        case 's':
        case 'S':
          break;
        default:
          return false; 
      }
    }
    return true;
  }
  
  public static String num2NEWS(int sur)
  {
    switch (sur)
    {
      default: return "xxxx (#0) "; 
      case 1: return "Nxxx (#1) "; 
      case 2: return "xExx (#2) "; 
      case 3: return "NExx (#3) "; 
      case 4: return "xxxS (#4) "; 
      case 5: return "NxxS (#5) "; 
      case 6: return "xExS (#6) "; 
      case 7: return "NExS (#7) "; 
      case 8: return "xxWx (#8) "; 
      case 9: return "NxWx (#9) "; 
      case 10: return "xEWx (#10)"; 
      case 11: return "NEWx (#11)"; 
      case 12: return "xxWS (#12)"; 
      case 13: return "NxWS (#13)"; 
      case 14: return "xEWS (#14)"; 
      case 15: return "NEWS (#15)"; 
    }
  }
    
    
    public static void parseRules(String infilename)
    {
      FileReader f;  
      BufferedReader b;
      StreamTokenizer s;
        
    // they have to define what happens in State 0 !
    int tt = 0;
    int cs=0; // current state, used to index the above array
    int sur=0; // surroundings
    String surNEWS = "";  // surroundings in NEWS format
    int move=0; // the move
    int ns=0; // the new state
    int markerin=Marker.NO; // whether this transition relies on a marker
    int markerout=Marker.NONE; // whether the output does anything w/ a marker
    
    try
    { 
      f = new java.io.FileReader(infilename);
      b = new java.io.BufferedReader(f);
      s = new java.io.StreamTokenizer(b);
      s.eolIsSignificant(true);
      s.lowerCaseMode(true);
      s.wordChars('!','~');
      s.ordinaryChar('/');
      s.ordinaryChar('#');
      s.commentChar('#');
      s.parseNumbers();
      //s.ordinaryChar('-');
      s.ordinaryChar('>');
      s.slashStarComments(true);
      s.slashSlashComments(true);

        while ( (tt=s.nextToken()) != java.io.StreamTokenizer.TT_EOF )
        {
          
          if ( tt == java.io.StreamTokenizer.TT_EOL )
          {
            continue;
          }
          else
          {
            if ( tt == java.io.StreamTokenizer.TT_WORD &&
                 s.sval.equals("*") ) // must be an "all current states rule"
            {
              cs = -1;
            }
              else if ( tt == java.io.StreamTokenizer.TT_NUMBER )
              {
                //H.out.println("initial state: " + s.nval);
                cs = (int)s.nval;
                if (cs < 0 || cs >= MAXSTATE)
                {
                  H.out.println("Error on line " + s.lineno() +
                   " : the starting state is not between 0 and " +(MAXSTATE-1)+ "\n");
                  System.exit(0);
                }
                // default next state = current state
                ns = cs;
                // integer -- no checking for marker here
                markerin = Marker.NO;
              }
              else if ( tt == java.io.StreamTokenizer.TT_WORD )
              {
                char c = s.sval.charAt(0);
                switch (c)
                {
                  case 'm': markerin = Marker.YES; break;
                  default:
                  {
                  H.out.println("Error on line " + s.lineno() +
                   " : the word to indicate the presence of a marker\n" +
                   "    does NOT begin with an m.\n");
                  System.exit(0);
                  }
                }
                
                // now get the first integer for the state
                tt=s.nextToken();
                
                if ( tt == java.io.StreamTokenizer.TT_WORD &&
                     s.sval.equals("*") ) // rule designating all states
                {
                  cs = -1;
                }
                else if ( tt == java.io.StreamTokenizer.TT_NUMBER )
                  {
                    //H.out.println("state: " + s.nval);
                    cs = (int)s.nval;
                    if (cs < 0 || cs >= MAXSTATE)
                    {
                      H.out.println("Error on line " + s.lineno() +
                       " : the starting state is not between 0 and " +(MAXSTATE-1)+ "\n");
                      System.exit(0);
                    }
                    // default next state = current state
                    ns = cs;
                  }
                  else
                  {
                    H.out.println("Error on line " + s.lineno() +
                     " : it did not have an integer or '*' (for the current state)\n" +
                     "    after the marker indication (m).\n");
                    System.exit(0);
                  }
              }
              else
              {
                H.out.println("Error on line " + s.lineno() +
                 " : line did not start with an integer, a '*', or an 'm'.\n");
                System.exit(0);
              }
              
              tt=s.nextToken(); // get the surroundings
              
              if ( tt == java.io.StreamTokenizer.TT_NUMBER )
              {
                //H.out.println("number: " + s.nval);
                sur = (int)s.nval;
                if (sur < 0 || sur >= MAXSUR)
                {
                  H.out.println("Error on line " + s.lineno() +
                   " : the integer code for the surroundings\n  " +
                   "  is not between 0 and " +(MAXSUR-1)+ "\n");
                  System.exit(0);
                }
              }
              else if ( tt == java.io.StreamTokenizer.TT_WORD &&
                        s.sval.equals("*") )
              {
                // H.out.println("Found a '*' sur\n");
                sur = -1; // sentinel for default value
              }
              else if ( tt == java.io.StreamTokenizer.TT_WORD &&
                 s.sval.equals("x") )
              {
                // H.out.println("Found a 'x' sur\n");
                sur = 0;  // zero surroundings
              }
              else if ( tt == java.io.StreamTokenizer.TT_WORD )
              {
                // allow for NEWS specification of surroundings
                if (s.sval.length() > 4)  // need to have 1 per NEWS slot
                {
                  H.out.println("Error on line " + s.lineno() +
                     " : the surroundings " + s.sval + " is more than 4 characters long.\n");
                  System.exit(0);
                }
                // if it's exactly length 4, we have some
                // special checking to do if it uses an x or a *
                if (s.sval.length() == 4 && 
                    (s.sval.indexOf('x') > -1 ||
                     s.sval.indexOf('*') > -1))
                {   
                    // now, we know it's 4 chars long, make sure each is a * or x or the
                    // appropriate letter             
                    char c0 = s.sval.charAt(0); 
                    char c1 = s.sval.charAt(1);
                    char c2 = s.sval.charAt(2);
                    char c3 = s.sval.charAt(3);
                    if (c0 != 'n' && c0 != 'N' && c0 != '*' && c0 != 'x')
                    {
                      H.out.println("Error on line " + s.lineno() +
                         " : the first character of the 4-letter surroundings " + s.sval + 
                         "\n  is not an n, N, x, or *\n");
                      System.exit(0);
                    }
                    if (c1 != 'e' && c1 != 'E' && c1 != '*' && c1 != 'x')
                    {
                      H.out.println("Error on line " + s.lineno() +
                         " : the second character of the 4-letter surroundings " + s.sval + 
                         "\n  is not an e, E, x, or *\n");
                      System.exit(0);
                    }
                    if (c2 != 'w' && c2 != 'W' && c2 != '*' && c2 != 'x')
                    {
                      H.out.println("Error on line " + s.lineno() +
                         " : the third character of the 4-letter surroundings " + s.sval + 
                         "\n  is not a w, W, x, or *\n");
                      System.exit(0);
                    }
                    if (c3 != 's' && c3 != 'S' && c3 != '*' && c3 != 'x')
                    {
                      H.out.println("Error on line " + s.lineno() +
                         " : the fourth character of the 4-letter surroundings " + s.sval + 
                         "\n  is not an s, S, x, or *\n");
                      System.exit(0);
                    }
                    // if all are *'s, we make sur = -1 (the default flag)
                    // actually, no!  **** is different than * because
                    // **** forces all 16 possibilities to adhere to this rule
                    // * catches all remaining unspecified possibilities
                    
                    // we use a flag of -2 for the NEWS format
                    sur = -2;
                    surNEWS = s.sval;        
                    // H.out.println(" Found surNews of " + surNEWS + "\n"); 
                }// end of s.sval.length() == 4 exactly  with an 'x' or an '*'
                else if ( isCompass(s.sval) ) 
                  // length is less than 4 and all are compass directions NEWS
                {
                  // set sur appropriately
                  sur = 0;
                  for (int i=0 ; i<s.sval.length() ; ++i)
                  {
                    char c = s.sval.charAt(i);
                    if (c == 'n' || c == 'N') { sur = sur | 1; }
                    if (c == 'e' || c == 'E') { sur = sur | 2; }
                    if (c == 'w' || c == 'W') { sur = sur | 8; }
                    if (c == 's' || c == 'S') { sur = sur | 4; }
                  }
                  if (sur == 0)
                  {
                    H.out.println("Error on line " + s.lineno() +
                     " : an valid surroundings was expected\n" +
                     "    but " + s.sval + " was found.\n");
                    System.exit(0);
                  }
                }
                else
                {     
                    H.out.println("Error on line " + s.lineno() +
                     " : an valid surroundings was expected\n" +
                     "    but " + s.sval + " was found.\n");
                    System.exit(0);
                }
              }
              else
              {
                H.out.println("Error on line " + s.lineno() +
                 " : an valid surroundings was expected\n" +
                 "    but " + s.sval + " was found.\n");
                System.exit(0);
              }
              
              tt=s.nextToken();
              
              if ( tt == java.io.StreamTokenizer.TT_WORD &&
                   s.sval.equals("->") )
              {
                ;
              }
              else if ( tt == '-' )
              {
                tt = s.nextToken();
                if ( tt == '>' )
                {
                  ;
                }
                else
                {
                  H.out.println("Error on line " + s.lineno() +
                    " : an arrow -> was expected, but not found.\n\n");
                  System.exit(0);
                }
              }
              else
              {
                /*
                H.out.println("tt is " + (int)(tt) );
                H.out.println("ttW is " + (int)(java.io.StreamTokenizer.TT_WORD) );
                H.out.println("ttN is " + (int)(java.io.StreamTokenizer.TT_NUMBER) ); */
                H.out.println("Error on line " + s.lineno() +
                 " : an arrow -> was expected, but not found.  " + s.sval + " was found\n\n");
                System.exit(0);
              }
                
              tt=s.nextToken();
              
              if ( tt == java.io.StreamTokenizer.TT_WORD )
              {
                markerout = Marker.NONE;     // default marker value -- no marker command
                //H.out.println("word: " + s.sval);
                char c = s.sval.charAt(0);
                switch (c)
                {
                  case 'n': move = Move.N; break;
                  case 'e': move = Move.E; break;
                  case 'w': move = Move.W; break;
                  case 's': move = Move.S; break;
                  case 'x': move = Move.X; break;
                  case 'd': markerout = Marker.DROP; break;
                  case 'p': markerout = Marker.PICKUP; break;
                  default:
                  {
                  H.out.println("Error on line " + s.lineno() +
                   " : the word (after the arrow) indicating either \n       the robot's next move or" +
                   " a marker command (drop or pickup) does\n" +
                   "    NOT begin with n,s,e,w,d,p, or x.\n (But it should!)\n");
                  System.exit(0);
                  }
                }
                
                // check if it was a marker command
                
                if (markerout != Marker.NONE)
                {
                  // need to get the next move
                  tt=s.nextToken();
              
                      if ( tt == java.io.StreamTokenizer.TT_WORD )
                      {
                        
                        //H.out.println("word: " + s.sval);
                        c = s.sval.charAt(0);
                        switch (c)
                        {
                          case 'n': move = Move.N; break;
                          case 'e': move = Move.E; break;
                          case 'w': move = Move.W; break;
                          case 's': move = Move.S; break;
                          case 'x': move = Move.X; break;
                          default:
                          {
                          H.out.println("Error on line " + s.lineno() +
                           " : the word (after the marker command), indicating\n" +
                           "       the robot's next move, does" +
                           " NOT begin with n,s,e,w, or x.\n (But it should!)\n");
                          System.exit(0);
                          }
                        }
                      }
                      else
                      {
                        H.out.println("Error on line " + s.lineno() +
                                  " : the word to indicate the robot's next move\n" +
                                  "      (after the marker command) does " +
                                  "    NOT begin with a letter.\n");
                    System.exit(0);
                      }
                }
              }
              else
              {
                H.out.println("Error on line " + s.lineno() +
                 " : the word to indicate the robot's next move does\n" +
                 "    NOT begin with a letter.\n");
                System.exit(0);
              }
              
              tt=s.nextToken(); // get the next state (required)
              
              if ( tt == java.io.StreamTokenizer.TT_NUMBER )
              {
                //H.out.println("state: " + s.nval);
                ns = (int)s.nval;
                if (ns < 0 || ns >= MAXSTATE)
                {
                  H.out.println("Error on line " + s.lineno() +
                   " : the robot's next state is not between 0 and " +(MAXSTATE-1)+ "\n");
                  System.exit(0);
                }
              }
              else
              {
                H.out.println("Error on line " + s.lineno() +
                 " : the state that the robot will transition to\n" +
                 "      is not an integer or is not present.\n");
                System.exit(0);
              }
              
              tt=s.nextToken();
              
              if ( tt != java.io.StreamTokenizer.TT_EOL &&
                   tt != java.io.StreamTokenizer.TT_EOF )
              {
                H.out.println("Error on line " + s.lineno() +
                 " : I found additional text after the rule.\n");
                System.exit(0);
              }
            } // end of line
            
            /*
             * The rule should be parsed, now we'll do error checking
             *    and eventually get the rule into the right place
             */
             
             // H.out.println("Line number: " + (s.lineno()-1));
             
            /*
             * The above keeps track of what line we're looking at 
             */
             
          if (cs != -1)  // is this the default rule for all unspecified start states?
          {
            // is this state null ?
            if ( state[cs] == null )
            {
              state[cs] = new State();
            }
            
          // marker ?
          if ( markerin == Marker.NO )
          {
            // has this surroundings transition already been specified?
            // first, the default case
            if ( sur == -1 )
            {
              if ( state[cs].def != null )
                {
                  H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the default (*) action\n"+
                       "    for state #" + cs + " for a second time.\n\n" + 
                       "    The first definition was on line " + state[cs].def.linenum + ".\n");
                  System.exit(0);
                }
                else // OK, add this one
                {
                  // check if markerout is NOT Marker.PICKUP
                  // (you can Marker.DROP or Marker.NONE, but since
                  //  there's no marker you can't Marker.PICKUP!)
                  if ( markerout == Marker.PICKUP )
                  {
                    System.out.println("On line " + (s.lineno()-1) + 
                     " you have asked the robot to pickup a marker\n" + 
                     "      in an instance when there would be no marker to pick up.\n");
                    System.exit(0);
                  }
                  if ( markerout != Marker.DROP ) {markerout = Marker.NONE;}
                  state[cs].def = new Trans(ns,move,markerout,s.lineno()-1);
                }
              }
            else if ( sur == -2 ) // the case when the surroundings is specified in NEWS format
            {
              // it's already been checked for correct form
              // thus, we need only see which entries are '*'
              
              char c0 = surNEWS.charAt(0);
              char c1 = surNEWS.charAt(1);
              char c2 = surNEWS.charAt(2);
              char c3 = surNEWS.charAt(3);
              int nL,nH,eL,eH,wL,wH,sL,sH;
              
              if (c0 == '*') { nL = 0 ; nH = 1; }
              else if (c0 == 'x') { nL = 0 ; nH = 0; }
              else  { nL = 1 ; nH = 1; }
              
              if (c1 == '*') { eL = 0 ; eH = 2; }
              else if (c1 == 'x') { eL = 0 ; eH = 0; }
              else  { eL = 2 ; eH = 2; }
              
              if (c2 == '*') { wL = 0 ; wH = 8; }
              else if (c2 == 'x') { wL = 0 ; wH = 0; }
              else  { wL = 8 ; wH = 8; }
              
              if (c3 == '*') { sL = 0 ; sH = 4; }
              else if (c3 == 'x') { sL = 0 ; sH = 0; }
              else  { sL = 4 ; sH = 4; }
              
              for (int ni = nL ; ni<= nH ; ni += 1)
                for (int ei = eL ; ei<= eH ; ei += 2)
                  for (int wi = wL ; wi<= wH ; wi += 8)
                    for (int si = sL ; si<= sH ; si += 4)
                    {
                      sur = ni+ei+wi+si;
                        // borrowed from below
                        if ( state[cs].trans[sur] != null )
                        { 
                          H.out.println("Error on line " + (s.lineno()-1) +
                               " : you have defined the action\n"+
                               "    for state #" + cs + " and surroundings " +
                               num2NEWS(sur) + " for a second time.\n\n" +
                          "    The first definition was on line " + 
                          state[cs].trans[sur].linenum + ".\n");
                          System.exit(0);
                        }
                        else // OK, add this one
                        {
                          if ( markerout == Marker.PICKUP )
                          {
                            System.out.println("On line " + (s.lineno()-1) + 
                             " you have asked the robot to pickup a marker\n" + 
                             "      in an instance when there would be no marker to pick up.\n");
                            System.exit(0);
                          }
                          if ( markerout != Marker.DROP ) {markerout = Marker.NONE;}
                          state[cs].trans[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                        }
                    }
               sur = -2; // reset sur
            }
            else // the case when the surroundings is specified as a number
              {
                if ( state[cs].trans[sur] != null )
                { 
                  H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the action\n"+
                       "    for state #" + cs + " and surroundings " +
                       num2NEWS(sur) + " for a second time.\n\n" +
                  "    The first definition was on line " + 
                  state[cs].trans[sur].linenum + ".\n");
                  System.exit(0);
                }
                else // OK, add this one
                {
                  if ( markerout == Marker.PICKUP )
                  {
                    System.out.println("On line " + (s.lineno()-1) + 
                     " you have asked the robot to pickup a marker\n" + 
                     "      in an instance when there would be no marker to pick up.\n");
                    System.exit(0);
                  }
                  if ( markerout != Marker.DROP ) {markerout = Marker.NONE;}
                  state[cs].trans[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                }
              }   
            } // end of if markerin == Marker.NO
            else //     if markerin == Marker.YES
            {
              // has this surroundings transition already been specified?
            // first, the default case
            if ( sur == -1 )
            {
              if ( state[cs].defMarker != null )
                {
                  H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the default (*) action\n"+
                       "    for state " + cs + " (when finding a marker) for a second time.\n\n"+
                  "    The first definition was on line " + 
                  state[cs].defMarker.linenum + ".\n");
                  System.exit(0);
                }
                else // OK, add this one
                {
                  // could check to be sure the marker command is not Marker.DROP
                  state[cs].defMarker = new Trans(ns,move,markerout,s.lineno()-1);
                }
              }
            else if ( sur == -2 ) // the case when the surroundings is specified in NEWS format
            {
              // it's already been checked for correct form
              // thus, we need only see which entries are '*'
              
              char c0 = surNEWS.charAt(0);
              char c1 = surNEWS.charAt(1);
              char c2 = surNEWS.charAt(2);
              char c3 = surNEWS.charAt(3);
              int nL,nH,eL,eH,wL,wH,sL,sH;
              
              if (c0 == '*') { nL = 0 ; nH = 1; }
              else if (c0 == 'x') { nL = 0 ; nH = 0; }
              else  { nL = 1 ; nH = 1; }
              
              if (c1 == '*') { eL = 0 ; eH = 2; }
              else if (c1 == 'x') { eL = 0 ; eH = 0; }
              else  { eL = 2 ; eH = 2; }
              
              if (c2 == '*') { wL = 0 ; wH = 8; }
              else if (c2 == 'x') { wL = 0 ; wH = 0; }
              else  { wL = 8 ; wH = 8; }
              
              if (c3 == '*') { sL = 0 ; sH = 4; }
              else if (c3 == 'x') { sL = 0 ; sH = 0; }
              else  { sL = 4 ; sH = 4; }
              
              for (int ni = nL ; ni<= nH ; ni += 1)
                for (int ei = eL ; ei<= eH ; ei += 2)
                  for (int wi = wL ; wi<= wH ; wi += 8)
                    for (int si = sL ; si<= sH ; si += 4)
                    {
                      sur = ni+ei+wi+si;
                        // borrowed from below
                        if ( state[cs].transMarker[sur] != null )
                        { 
                          H.out.println("Error on line " + (s.lineno()-1) +
                               " : you have defined the action\n"+
                               "    for state #" + cs + " and surroundings " +
                               num2NEWS(sur) + " (with a marker) for a second time.\n\n" +
                          "    The first definition was on line " + 
                          state[cs].transMarker[sur].linenum + ".\n");
                          System.exit(0);
                        }
                        else // OK, add this one
                        {
                          state[cs].transMarker[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                        }
                    }
               sur = -2; // reset sur
            }
              else // same check for nondefault cases
              {
                if ( state[cs].transMarker[sur] != null )
                { 
                  H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the action\n"+
                       "    for state " + cs + " and surroundings " +
                       num2NEWS(sur) + " (with a marker) for a second time.\n\n" +
                  "    The first definition was on line " + 
                  state[cs].transMarker[sur].linenum + ".\n");
                  System.exit(0);
                }
                else // OK, add this one
                {
                  // could check to be sure the marker command is not Marker.DROP
                  state[cs].transMarker[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                }
              }   
            } // end of if markerin == Marker.YES (really an "else")
            
            } // end of if (cs != -1), i.e., the above code handles explicit current states
            else
            {     // if cs == -1, we are handling * as the current state
              // check for markerin status
              if (markerin == Marker.YES)
              {
                if (sur == -1)
                {
                  if (State.defAllMarker.def != null)
                  {
                    H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the action upon seeing a marker (m)\n"+
                       "    for the default current state (*)\n" + " and default surroundings (*)" +
                       " for a second time.\n\n" +
                    "    The first definition was on line " + 
                    State.defAllMarker.def.linenum + ".\n");
                    System.exit(0);
                  }
                  else
                  {
                    State.defAllMarker.def = new Trans(ns,move,markerout,s.lineno()-1);
                  }
                }
            else if ( sur == -2 ) // the case when the surroundings is specified in NEWS format
            {
              // it's already been checked for correct form
              // thus, we need only see which entries are '*'
              
              char c0 = surNEWS.charAt(0);
              char c1 = surNEWS.charAt(1);
              char c2 = surNEWS.charAt(2);
              char c3 = surNEWS.charAt(3);
              int nL,nH,eL,eH,wL,wH,sL,sH;
              
              if (c0 == '*') { nL = 0 ; nH = 1; }
              else if (c0 == 'x') { nL = 0 ; nH = 0; }
              else  { nL = 1 ; nH = 1; }
              
              if (c1 == '*') { eL = 0 ; eH = 2; }
              else if (c1 == 'x') { eL = 0 ; eH = 0; }
              else  { eL = 2 ; eH = 2; }
              
              if (c2 == '*') { wL = 0 ; wH = 8; }
              else if (c2 == 'x') { wL = 0 ; wH = 0; }
              else  { wL = 8 ; wH = 8; }
              
              if (c3 == '*') { sL = 0 ; sH = 4; }
              else if (c3 == 'x') { sL = 0 ; sH = 0; }
              else  { sL = 4 ; sH = 4; }
              
              for (int ni = nL ; ni<= nH ; ni += 1)
                for (int ei = eL ; ei<= eH ; ei += 2)
                  for (int wi = wL ; wi<= wH ; wi += 8)
                    for (int si = sL ; si<= sH ; si += 4)
                    {
                      sur = ni+ei+wi+si;
                        // borrowed from below
                        if ( State.defAllMarker.trans[sur] != null )
                        { 
                           H.out.println("Error on line " + (s.lineno()-1) +
                           " : you have defined the action upon seeing a marker (m)\n"+
                           "    for the default current state (*)\n" + "    and surroundings #" +
                           sur + " for a second time.\n\n" +
                           "    The first definition was on line " + 
                           State.defAllMarker.trans[sur].linenum + ".\n");
                           System.exit(0);
                        }
                        else // OK, add this one
                        {
                          State.defAllMarker.trans[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                        }
                    }
               sur = -2; // reset sur
               }
                else // sur != -1 or -2
                {
                  if (State.defAllMarker.trans[sur] != null)
                  {
                    H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the action upon seeing a marker (m)\n"+
                       "    for the default current state (*)\n" + "    and surroundings #" +
                       sur + " for a second time.\n\n" +
                       "    The first definition was on line " + 
                       State.defAllMarker.trans[sur].linenum + ".\n");
                    System.exit(0);
                  }
                  else
                  {
                    State.defAllMarker.trans[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                  }
                }
                
              }
              else // markerin == Marker.NO
              {
              
                if (sur == -1)
                {
                  if (State.defAll.def != null)
                  {
                    H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the action\n"+
                       "    for the default current state (*)\n" + " and default surroundings (*)" +
                       " twice.\n\n" +
                       "    The first definition was on line " + 
                       State.defAll.def.linenum + ".\n");
                    System.exit(0);
                  }
                  else
                  {
                    State.defAll.def = new Trans(ns,move,markerout,s.lineno()-1);
                  }
                }
                else if ( sur == -2 ) // the case when the surroundings is specified in NEWS format
            {
              // it's already been checked for correct form
              // thus, we need only see which entries are '*'
              
              char c0 = surNEWS.charAt(0);
              char c1 = surNEWS.charAt(1);
              char c2 = surNEWS.charAt(2);
              char c3 = surNEWS.charAt(3);
              int nL,nH,eL,eH,wL,wH,sL,sH;
              
              if (c0 == '*') { nL = 0 ; nH = 1; }
              else if (c0 == 'x') { nL = 0 ; nH = 0; }
              else  { nL = 1 ; nH = 1; }
              
              if (c1 == '*') { eL = 0 ; eH = 2; }
              else if (c1 == 'x') { eL = 0 ; eH = 0; }
              else  { eL = 2 ; eH = 2; }
              
              if (c2 == '*') { wL = 0 ; wH = 8; }
              else if (c2 == 'x') { wL = 0 ; wH = 0; }
              else  { wL = 8 ; wH = 8; }
              
              if (c3 == '*') { sL = 0 ; sH = 4; }
              else if (c3 == 'x') { sL = 0 ; sH = 0; }
              else  { sL = 4 ; sH = 4; }
              
              for (int ni = nL ; ni<= nH ; ni += 1)
                for (int ei = eL ; ei<= eH ; ei += 2)
                  for (int wi = wL ; wi<= wH ; wi += 8)
                    for (int si = sL ; si<= sH ; si += 4)
                    {
                      sur = ni+ei+wi+si;
                        // borrowed from below
                        if (State.defAll.trans[sur] != null)
                          {
                            H.out.println("Error on line " + (s.lineno()-1) +
                               " : you have defined the action \n"+
                               "    for the default current state (*)\n" + "    and surroundings #" +
                               sur + " twice.\n\n" +
                               "    The first definition was on line " + 
                               State.defAll.trans[sur].linenum + ".\n");
                            System.exit(0);
                          }
                        else // OK, add this one
                        {
                          if ( markerout == Marker.PICKUP )
                          {
                            System.out.println("On line " + (s.lineno()-1) + 
                             " you have asked the robot to pickup a marker\n" + 
                             "      in an instance when there would be no marker to pick up.\n");
                            System.exit(0);
                          }
                          if ( markerout != Marker.DROP ) {markerout = Marker.NONE;}
                          State.defAll.trans[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                        }
                    }
               sur = -2; // reset sur
               }
                else // sur != -1 or -2
                {
                  if (State.defAll.trans[sur] != null)
                  {
                    H.out.println("Error on line " + (s.lineno()-1) +
                       " : you have defined the action \n"+
                       "    for the default current state (*)\n" + "    and surroundings #" +
                       sur + " twice.\n\n" +
                       "    The first definition was on line " + 
                       State.defAll.trans[sur].linenum + ".\n");
                    System.exit(0);
                  }
                  else
                  {
                    if ( markerout == Marker.PICKUP )
                      {
                        System.out.println("On line " + (s.lineno()-1) + 
                         " you have asked the robot to pickup a marker\n" + 
                         "      in an instance when there would be no marker to pick up.\n");
                        System.exit(0);
                      }
                      if ( markerout != Marker.DROP ) {markerout = Marker.NONE;}
                    State.defAll.trans[sur] = new Trans(ns,move,markerout,s.lineno()-1);
                  }
                }

              }

            }
            
             H.out.println("Line # " + (s.lineno()-1) + " is OK");
             
            /*
             * The above keeps track of what line we're looking at 
             */
            
            
        } // end of file getting parsed
        
        // check the rules that state zero is there and
        // that no state is transitioned to that does not exist!
        
        checkRules(state);
        
        H.pl("\nThe rules are as follows:\n");
        
        printRules(state);
        
        H.out.println("\nDone with parsing\n");
        
        /*****************************************************/
   
    } catch (Exception e) {
        H.pl("Problem with picobot.txt!");
        e.printStackTrace();
    }


    } // end of parsing
    
    public boolean isWater(int r, int c)
    {
        return grid[r][c].getColor() == waterColor;
    }
    
    public int getSurroundings()
      {
        int s = 0;
    
        // bit 3 - west
        if (isWater(roboty,robotx-1)) { s |= 1<<3; }
        // bit 2 - south
        if (isWater(roboty+1,robotx)) { s |= 1<<2; }
        // bit 1 - east
        if (isWater(roboty,robotx+1)) { s |= 1<<1; }
        // bit 0 - north
        if (isWater(roboty-1,robotx)) { s |= 1; }

        return s; 
      }

    
    public int getNextMove()
    {
        // update surroundings
      int S = getSurroundings();
      boolean hasPebble = grid[roboty][robotx].getPebble();
      int rx = robotx;
      int ry = roboty;
      int nextMove = Move.X;     // no move
      int newState = statenow;   // current state is "statenow"
      int linenum = 0;           // linenumber for rule
      boolean changePebble = false;
      String ruleUsed = "no rule available";
      
      // for an explanation of the method "cb", see "computeBuffer", below
      
      if (!hasPebble)
      {
          if (state[statenow].trans[S] != null)
          {
            nextMove = state[statenow].trans[S].move;
            newState = state[statenow].trans[S].ns;
            linenum = state[statenow].trans[S].linenum;
            changePebble = (state[statenow].trans[S].marker == Marker.DROP);
            ruleUsed = "(  " + cb(statenow) + " " + cb(S) + "  ->  " + 
                       Marker.print(state[statenow].trans[S].marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (state[statenow].def != null)
          {
            nextMove = state[statenow].def.move;
            newState = state[statenow].def.ns;
            linenum = state[statenow].def.linenum;
            changePebble = (state[statenow].def.marker == Marker.DROP);
            ruleUsed = "(  " + cb(statenow) + "  *  ->  " + 
                       Marker.print(state[statenow].def.marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" + 
                       "   from line # " + linenum;
          }
          else if (State.defAll != null && State.defAll.trans[S] != null)
          {
            nextMove = State.defAll.trans[S].move;
            newState = State.defAll.trans[S].ns;
            linenum = State.defAll.trans[S].linenum;
            changePebble = (State.defAll.trans[S].marker == Marker.DROP);
            ruleUsed = "(   *" + " " + cb(S) + "  ->  " + 
                       Marker.print(State.defAll.trans[S].marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (State.defAll != null && State.defAll.def != null)
          {
            nextMove = State.defAll.def.move;
            newState = State.defAll.def.ns;
            linenum = State.defAll.def.linenum;
            changePebble = (State.defAll.def.marker == Marker.DROP);
            ruleUsed = "(   *" + "  *  ->  " + 
                       Marker.print(State.defAll.def.marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else
          {
            H.out.println("There is no rule to handle this situation:");
            H.out.println("  surroundings = " + S);
            H.out.println("  current state = " + statenow);
            nextMove = Move.X;
            newState = statenow;
            changePebble = false;
            ruleUsed = "none available";
          }
        }
        else // there _IS_ a marker (pebble)
        {
          if (state[statenow].transMarker[S] != null)
          {
            nextMove = state[statenow].transMarker[S].move;
            newState = state[statenow].transMarker[S].ns;
            linenum = state[statenow].transMarker[S].linenum;
            changePebble = (state[statenow].transMarker[S].marker == Marker.PICKUP);
            ruleUsed = "(m " + cb(statenow) + " " + cb(S) + "  ->  " + 
                       Marker.print(state[statenow].transMarker[S].marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (state[statenow].defMarker != null)
          {
            nextMove = state[statenow].defMarker.move;
            newState = state[statenow].defMarker.ns;
            linenum = state[statenow].defMarker.linenum;
            changePebble = (state[statenow].defMarker.marker == Marker.PICKUP);
            ruleUsed = "(m " + cb(statenow) + "  *  ->  " + 
                       Marker.print(state[statenow].defMarker.marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (State.defAllMarker != null && State.defAllMarker.trans[S] != null)
          {
            nextMove = State.defAllMarker.trans[S].move;
            newState = State.defAllMarker.trans[S].ns;
            linenum = State.defAllMarker.trans[S].linenum;
            changePebble = (State.defAllMarker.trans[S].marker == Marker.PICKUP);
            ruleUsed = "(m  *" + " " + cb(S) + "  ->  " + 
                       Marker.print(State.defAllMarker.trans[S].marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (State.defAllMarker != null && State.defAllMarker.def != null)
          {
            nextMove = State.defAllMarker.def.move;
            newState = State.defAllMarker.def.ns;
            linenum = State.defAllMarker.def.linenum;
            changePebble = (State.defAllMarker.def.marker == Marker.PICKUP);
            ruleUsed = "(m  *" + "  *  ->  " + 
                       Marker.print(State.defAllMarker.def.marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (state[statenow].trans[S] != null)
          {
            nextMove = state[statenow].trans[S].move;
            newState = state[statenow].trans[S].ns;
            linenum = state[statenow].trans[S].linenum;
            changePebble = (state[statenow].trans[S].marker == Marker.PICKUP);
            ruleUsed = "(  " + cb(statenow) + " " + cb(S) + "  ->  " + 
                       Marker.print(state[statenow].trans[S].marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (state[statenow].def != null)
          {
            nextMove = state[statenow].def.move;
            newState = state[statenow].def.ns;
            linenum = state[statenow].def.linenum;
            changePebble = (state[statenow].def.marker == Marker.PICKUP);
            ruleUsed = "(  " + cb(statenow) + "  *  ->  " + 
                       Marker.print(state[statenow].def.marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (State.defAll != null && State.defAll.trans[S] != null)
          {
            nextMove = State.defAll.trans[S].move;
            newState = State.defAll.trans[S].ns;
            linenum = State.defAll.trans[S].linenum;
            changePebble = (State.defAll.trans[S].marker == Marker.PICKUP);
            ruleUsed = "(   *" + " " + cb(S) + "  ->  " + 
                       Marker.print(State.defAll.trans[S].marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else if (State.defAll != null && State.defAll.def != null)
          {
            nextMove = State.defAll.def.move;
            newState = State.defAll.def.ns;
            linenum = State.defAll.def.linenum;
            changePebble = (State.defAll.def.marker == Marker.PICKUP);
            ruleUsed = "(   *" + "  *  ->  " + 
                       Marker.print(State.defAll.def.marker) + " " + 
                       Move.print(nextMove) + " " + cb(newState) + " )" +
                       "   from line # " + linenum;
          }
          else
          {
            H.out.println("There is no rule to handle this situation (with marker):");
            H.out.println("  surroundings = " + S);
            H.out.println("  current state = " + statenow);
            nextMove = Move.X;
            newState = statenow;
            changePebble = false;
            ruleUsed = "none available";
          }
        }
          
          
      /*
      String b = computeBuffer(rx,3); // b is the number of spaces
                                      // needed to add to rx to ensure
                                      // at least 3 total spaces
      
      
      H.out.print("Robot at (" + b + rx);
      b = computeBuffer(ry,3);
      H.out.print("," + b + ry + ")     " + "Rule used is  ");
      */
      H.out.println("  Rule applied:  " + ruleUsed);
     
      if (changePebble)     // if we need to update the marker
      {
        grid[roboty][robotx].setPebble(!hasPebble);
      }
      
      statenow = newState;
      return nextMove;
  }

    
    
    
} // end of Picobot class
	